/**
 * 
 */
package wj.demo.na;

import java.util.Iterator;

import org.geoserver.config.GeoServer;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.graph.build.feature.FeatureGraphGenerator;
import org.geotools.graph.build.line.LineStringGraphGenerator;
import org.geotools.graph.path.DijkstraShortestPathFinder;
import org.geotools.graph.path.Path;
import org.geotools.graph.structure.Edge;
import org.geotools.graph.structure.Graph;
import org.geotools.graph.structure.Node;
import org.geotools.graph.traverse.standard.DijkstraIterator.EdgeWeighter;
import org.opengis.feature.Feature;
import org.opengis.feature.simple.SimpleFeature;

import wj.demos.KVPRequest;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;

/**
 * @author Administrator
 * 
 */
public class NAService {

	private GeoServer geoServer;
	
	public NAService(GeoServer geoServer) {
		this.geoServer = geoServer;
	}

	public GetRouteTransform getRoute(KVPRequest request) throws Exception {
		//Set and prepare parameters
		Geometry[] stops = (Geometry[]) request.getAttribute("stops");
		Point originPoint = stops[0].getCentroid();
		Point destinationPoint = stops[1].getCentroid();
		String graphName = request.getAttribute("graph").toString();

		//Build the graph
		FeatureGraphGenerator featureGen = new FeatureGraphGenerator(
				new LineStringGraphGenerator()
				);
		FeatureCollection<?, ?> networkFC = geoServer.getCatalog()
				.getFeatureTypeByName(graphName).getFeatureSource(null, null)
				.getFeatures();
		FeatureIterator<?> iter = networkFC.features();
		try {
			while (iter.hasNext()) {
				Feature feature = iter.next();
				featureGen.add(feature);
			}
		} finally {
			iter.close();
		}
		
		Graph networkGraph = featureGen.getGraph();
		Node source = getNearestGraphNode(networkGraph, originPoint);
		
		// weight the edges of the graph using the distance of each linestring
		EdgeWeighter weighter = new EdgeWeighter() {
			public double getWeight(org.geotools.graph.structure.Edge e) {
				SimpleFeature aLineString = (SimpleFeature) e.getObject();
				Geometry geom = (Geometry) aLineString.getDefaultGeometry();
				return geom.getLength();
			}
		};
		
		// calculate the cost(distance) of each graph node to the node closest
		// to the origin
		DijkstraShortestPathFinder dspf = new DijkstraShortestPathFinder(networkGraph, source, weighter);
		dspf.calculate();		
 
		//Get the route
		Node destination = getNearestGraphNode(networkGraph, destinationPoint);
		
		Path path = dspf.getPath(destination);
		if (path == null)
			throw new Exception(
					"Could not calculate the route to this destination"
							+ ". There is probably a problem with the network dataset (dangles etc...");

		Geometry route = null;
		
		Node previous = null, node = null;
		for (Iterator<?> ritr = path.riterator(); ritr.hasNext();) {
			node = (Node) ritr.next();
			if (previous != null) {
				// Adds the resulting edge into the vector
				Edge eg = node.getEdge(previous);
				SimpleFeature feature = (SimpleFeature) eg.getObject();
				Geometry geom = (Geometry) feature.getDefaultGeometry();
				if (route == null) {
					route = geom;
				} else {
					route = route.union(geom);
				}
			}
			previous = node;
		}		
		
		return new GetRouteTransform(route);
	}

	private Node getNearestGraphNode(Graph g, Point pointy) {
		double dist = 0;
		Node nearestNode = null;
		for (Object o : g.getNodes()) {
			Node n = (Node) o;
			Point p = ((Point) n.getObject());
			double newdist = calculateEuclideanDistance(pointy.getX(),
					pointy.getY(), p.getCoordinate().x, p.getCoordinate().y);
			
			if(nearestNode == null || newdist < dist){
				dist = newdist;
				nearestNode = n;
			}
		}
		
		// System.out.println("StartKoordinate X: " +
		// ((Point)nearestStart.getObject()).getCoordinate().x);
		// System.out.println("StartKoordinate Y: " +
		// ((Point)nearestStart.getObject()).getCoordinate().y);
		// System.out.println("Distant: " + dist);

		return nearestNode;
	}

	private double calculateEuclideanDistance(double xOrig, double yOrig,
			double xDest, double yDest) {
		double distance = Math.sqrt((xDest - xOrig) * (xDest - xOrig)
				+ (yDest - yOrig) * (yDest - yOrig));
		return distance;
	}

}
